
OPTION PROLOGUE:none
OPTION EPILOGUE:none

SDWtoASCII      PROTO
SQWtoASCII      PROTO
QWtoASCII       PROTO
QWtoAconvert    PROTO

; ------------------------------------------------------------------------------

align 16

SDWtoASCII    proc

 comment    * ------------------------------------------------------------
            Convert signed DWORD to ASCII string by IanB.

            On entry:
                EAX = signed value to convert
                ECX = pointer to output string buffer (11 chars minimum)

            Falls through to QWtoAconvert routine.
            EBX/EBP/EDI/ESI preserved.
            ------------------------------------------------------------ *

        push    ebx
        xor     edx, edx                    ; zero EDX
        push    esi
        push    edi
        push    ecx                         ; save pointer
        add     eax, edx                    ; check for sign in top bit
        jns     QWtoAconvert

        neg     eax
        mov     BYTE PTR [ecx], "-"
        add     ecx, 1
        jmp     QWtoAconvert
        
SDWtoASCII    endp

align 16

SQWtoASCII    proc

 comment    * ------------------------------------------------------------
            Convert signed QWORD to ASCII string by IanB.

            On entry:
                EAX/EDX = signed value to convert Lo/Hi
                ECX = pointer to output string buffer (21 chars minimum)

            Falls through to QWtoAconvert routine.
            EBX/EBP/EDI/ESI preserved.
            ------------------------------------------------------------ *

        push    ebx
        xor     ebx, ebx                    ; zero EBX
        push    esi
        push    edi
        push    ecx                         ; save pointer
        add     edx, ebx                    ; check for sign in top bit
        jns     QWtoAconvert

        add     ebx, 1
        not     eax
        not     edx
        mov     BYTE PTR [ecx], "-"
        add     ecx, ebx
        add     eax, ebx
        jc      @F
        xor     ebx, ebx
@@:
        add     edx, ebx
        jmp     QWtoAconvert
        
SQWtoASCII    endp

align 16

QWtoASCII    proc

 comment    * ------------------------------------------------------------
            Convert positive unsigned QWORD to ASCII string by IanB.
            For unsigned DWORD only, just call with EDX = 0.

            On entry:
                EAX/EDX = unsigned value to convert Lo/Hi
                ECX = pointer to output string buffer (21 chars minimum)

            Falls through to QW2Aconvert routine.
            EBX/EBP/EDI/ESI preserved.
            ------------------------------------------------------------ *

        push    ebx
        push    esi
        push    edi
        push    ecx                         ; save pointer

QWtoASCII    endp

QWtoAconvert    proc

 comment    * ------------------------------------------------------------
            Work routine for QWORD/DWORD to ASCII conversion by IanB.

            Alignment should be good with 4 ops in previous procedure.
            If OPTION NOSCOPED is used in this module, this procedure can
            be merged into the preceding QWtoASCII one and QWtoAconvert
            left at the top as a label for the jumps from SQW/SDWtoASCII.

            On entry:
                EAX/EDX = unsigned value to convert Lo/Hi
                ECX = pointer to output string buffer,
                      neg sign may already be written if signed value
                EBX/ESI/EDI/original string pointer pushed on stack

            On exit:
                EAX = count of chars written
                ECX = pointer to null-termination byte in output string
                      in case we want to add more chars afterwards
                EDX = original pointer to string buffer
                string buffer contains the null-terminated ASCII string
            EBX/EBP/EDI/ESI preserved
            ------------------------------------------------------------ *

        mov     esi, edx                    ; hold good values in EDI/ESI
        mov     edi, eax
        mov     bl, "0"
        test    edx, edx                    ; can we skip the slow top DWORD stuff
        jz      TestLowDWORD

        mov     bh, bl                      ; keep for convenience
        cmp     esi, 2H
        jb      Digit09
        jz      Digit10a
        cmp     edx, 17H
        jb      Digit10
        jz      Digit11a
        cmp     esi, 0e8H
        jb      Digit11
        jz      Digit12a
        cmp     edx, 918H
        jb      Digit12
        jz      Digit13a
        cmp     esi, 5af3H
        jb      Digit13
        jz      Digit14a
        cmp     edx, 38d7eH
        jb      Digit14
        jz      Digit15a
        cmp     esi, 2386f2H
        jb      Digit15
        jz      Digit16a
        cmp     edx, 1634578H
        jb      Digit16
        jz      Digit17a
        cmp     esi, 0de0b6b3H
        jb      Digit17
        jz      Digit18a
        cmp     edx, 8ac72304H
        jb      Digit18
        ja      Digit19
        cmp     eax, 89e80000H
        jb      Digit18

Digit19:
        sub     eax, 89e80000H              ; 10,000,000,000,000,000,000
        sbb     edx, 8ac72304H
        mov     edi, eax                    ; copy good values
        mov     esi, edx
        mov     BYTE PTR [ecx], "1"         ; max value of 19th digit is 1
        add     ecx, 1                      ; so no need to loop
        jmp     Digit18

Digit18a:
        cmp     eax, 0a7640000H
        jb      Digit17

Digit18:
        cmp     edx, 45639183H
        jb      @F
        sub     eax, 44f40000H              ; 5,000,000,000,000,000,000
        sbb     edx, 45639182H
        add     bl, 5
        mov     edi, eax                    ; save good values
        mov     esi, edx
@@:
        sub     eax, 0a7640000H             ; 1,000,000,000,000,000,000
        sbb     edx, 0de0b6b3H
        jb      @F
        add     bl, 1
        mov     edi, eax                    ; save good values
        mov     esi, edx
        jmp     @B
@@:
        mov     eax, edi                    ; copy good values to test
        mov     edx, esi
        mov     [ecx], bl
        mov     bl, bh
        add     ecx, 1
        jmp     Digit17

Digit17a:
        cmp     eax, 5d8a0000H
        jb      Digit16

Digit17:
        cmp     edx, 6f05b5aH
        jb      @F
        sub     eax, 0d3b20000H             ; 500,000,000,000,000,000
        sbb     edx, 6f05b59H
        add     bl, 5
        mov     edi, eax                    ; save good values
        mov     esi, edx
@@:
        sub     eax, 5d8a0000H              ; 100,000,000,000,000,000
        sbb     edx, 1634578H
        jb      @F
        add     bl, 1
        mov     edi, eax                    ; save good values
        mov     esi, edx
        jmp     @B
@@:
        mov     eax, edi                    ; copy good values to test
        mov     edx, esi
        mov     [ecx], bl
        mov     bl, bh
        add     ecx, 1
        jmp     Digit16

Digit16a:
        cmp     eax, 6fc10000H
        jb      Digit15

Digit16:
        cmp     edx, 0b1a2bdH
        jb      @F
        sub     eax, 2ec50000H              ; 50,000,000,000,000,000
        sbb     edx, 0b1a2bcH
        add     bl, 5
        mov     edi, eax                    ; save good values
        mov     esi, edx
@@:
        sub     eax, 6fc10000H              ; 10,000,000,000,000,000
        sbb     edx, 2386f2H
        jb      @F
        add     bl, 1
        mov     edi, eax                    ; save good values
        mov     esi, edx
        jmp     @B
@@:
        mov     eax, edi                    ; copy good values to test
        mov     edx, esi
        mov     [ecx], bl
        mov     bl, bh
        add     ecx, 1
        jmp     Digit15

Digit15a:
        cmp     eax, 0a4c68000H
        jb      Digit14

Digit15:
        cmp     edx, 11c37aH
        jb      @F
        sub     eax, 37e08000H              ; 5,000,000,000,000,000
        sbb     edx, 11c379H
        add     bl, 5
        mov     edi, eax                    ; save good values
        mov     esi, edx
@@:
        sub     eax, 0a4c68000H             ; 1,000,000,000,000,000
        sbb     edx, 38d7eH
        jb      @F
        add     bl, 1
        mov     edi, eax                    ; save good values
        mov     esi, edx
        jmp     @B
@@:
        mov     eax, edi                    ; copy good values to test
        mov     edx, esi
        mov     [ecx], bl
        mov     bl, bh
        add     ecx, 1
        jmp     Digit14

Digit14a:
        cmp     eax, 107a4000H
        jb      Digit13

Digit14:
        cmp     edx, 1c6c0H
        jb      @F
        sub     eax, 52634000H              ; 500,000,000,000,000
        sbb     edx, 1c6bfH
        add     bl, 5
        mov     edi, eax                    ; save good values
        mov     esi, edx
@@:
        sub     eax, 107a4000H              ; 100,000,000,000,000
        sbb     edx, 5af3H
        jb      @F
        add     bl, 1
        mov     edi, eax                    ; save good values
        mov     esi, edx
        jmp     @B
@@:
        mov     eax, edi                    ; copy good values to test
        mov     edx, esi
        mov     [ecx], bl
        mov     bl, bh
        add     ecx, 1
        jmp     Digit13

Digit13a:
        cmp     eax, 4e72a000H
        jb      Digit12

Digit13:
        cmp     edx, 2d7aH
        jb      @F
        sub     eax, 883d2000H              ; 50,000,000,000,000
        sbb     edx, 2d79H
        add     bl, 5
        mov     edi, eax                    ; save good values
        mov     esi, edx
@@:
        sub     eax, 4e72a000H              ; 10,000,000,000,000
        sbb     edx, 918H
        jb      @F
        add     bl, 1
        mov     edi, eax                    ; save good values
        mov     esi, edx
        jmp     @B
@@:
        mov     eax, edi                    ; copy good values to test
        mov     edx, esi
        mov     [ecx], bl
        mov     bl, bh
        add     ecx, 1
        jmp     Digit12

Digit12a:
        cmp     eax, 0d4a51000H
        jb      Digit11

Digit12:
        cmp     edx, 48dH
        jb      @F
        sub     eax, 27395000H              ; 5,000,000,000,000
        sbb     edx, 48cH
        add     bl, 5
        mov     edi, eax                    ; save good values
        mov     esi, edx
@@:
        sub     eax, 0d4a51000H             ; 1,000,000,000,000
        sbb     edx, 0e8H
        jb      @F
        add     bl, 1
        mov     edi, eax                    ; save good values
        mov     esi, edx
        jmp     @B
@@:
        mov     eax, edi                    ; copy good values to test
        mov     edx, esi
        mov     [ecx], bl
        mov     bl, bh
        add     ecx, 1
        jmp     Digit11

Digit11a:
        cmp     eax, 4876e800H
        jb      Digit10

Digit11:
        cmp     edx, 75H
        jb      @F
        sub     eax, 6a528800H              ; 500,000,000,000
        sbb     edx, 74H
        add     bl, 5
        mov     edi, eax                    ; save good values
        mov     esi, edx
@@:
        sub     eax, 4876e800H              ; 100,000,000,000
        sbb     edx, 17H
        jb      @F
        add     bl, 1
        mov     edi, eax                    ; save good values
        mov     esi, edx
        jmp     @B
@@:
        mov     eax, edi                    ; copy good values to test
        mov     edx, esi
        mov     [ecx], bl
        mov     bl, bh
        add     ecx, 1
        jmp     Digit10

Digit10a:
        cmp     eax, 540be400H
        jb      Digit09

Digit10:
        cmp     edx, 0cH
        jb      @F
        sub     eax, 0a43b7400H             ; 50,000,000,000
        sbb     edx, 0bH
        add     bl, 5
        mov     edi, eax                    ; save good values
        mov     esi, edx
@@:
        sub     eax, 540be400H              ; 10,000,000,000
        sbb     edx, 2H
        jb      @F
        add     bl, 1
        mov     edi, eax                    ; save good values
        mov     esi, edx
        jmp     @B
@@:
        mov     eax, edi                    ; copy good value to test
        mov     [ecx], bl
        mov     bl, bh
        add     ecx, 1

Digit09:
        mov     dx, 0ff01H                  ; convenience number
        xor     edi, edi                    ; zero EDI
        sub     bl, dl                      ; fix for following loop
        test    esi, esi
        jz      Digit09a
        sub     eax, 0ee6b2800H             ; 4,000,000,000
        sbb     esi, edi
        add     bl, 4
        test    esi, esi
        jz      Digit09a
        sub     eax, 0ee6b2800H             ; 4,000,000,000
        sbb     esi, edi
        add     bl, 4
        jmp     Digit09a                    ; high DWORD cleared

TestLowDWORD:
        mov     dx, 0ff01H                  ; convenience number
        cmp     eax, 10
        jb      Digit00
        cmp     edi, 100
        jb      Digit01
        cmp     eax, 1000
        jb      Digit02
        cmp     edi, 10000
        jb      Digit03
        cmp     eax, 100000
        jb      Digit04
        cmp     edi, 1000000
        jb      Digit05
        cmp     eax, 10000000
        jb      Digit06
        cmp     edi, 100000000
        jb      Digit07
        sub     eax, 1000000000
        jb      Digit08

Digit09a:
        mov     edi, eax                    ; save good value, no high DWORD
        add     bl, dl
        sub     eax, 1000000000
        jnb     Digit09a

        mov     [ecx], bl
        add     ecx, 1

Digit08:
        mov     eax, (-900000000)
        mov     esi, edi                    ; copy good value to test
        mov     bx, "90"
        add     eax, edi                    ; check if digit = 9
        jns     FromTop08
@@:
        sub     esi, 100000000
        jb      @F
        add     bx, dx                      ; BL + 1, BH - 1
        mov     edi, esi                    ; save good value
        add     eax, 100000000
        js      @B

FromTop08:
        mov     bl, bh
        mov     edi, eax
@@:
        mov     [ecx], bl
        add     ecx, 1

Digit07:
        mov     eax, (-90000000)
        mov     esi, edi                    ; copy good value to test
        mov     bx, "90"
        add     eax, edi                    ; check if digit = 9
        jns     FromTop07
@@:
        sub     esi, 10000000
        jb      @F
        add     bx, dx                      ; BL + 1, BH - 1
        mov     edi, esi                    ; save good value
        add     eax, 10000000
        js      @B

FromTop07:
        mov     bl, bh
        mov     edi, eax
@@:
        mov     [ecx], bl
        add     ecx, 1

Digit06:
        mov     eax, (-9000000)
        mov     esi, edi                    ; copy good value to test
        mov     bx, "90"
        add     eax, edi                    ; check if digit = 9
        jns     FromTop06
@@:
        sub     esi, 1000000
        jb      @F
        add     bx, dx                      ; BL + 1, BH - 1
        mov     edi, esi                    ; save good value
        add     eax, 1000000
        js      @B

FromTop06:
        mov     bl, bh
        mov     edi, eax
@@:
        mov     [ecx], bl
        add     ecx, 1

Digit05:
        mov     eax, (-900000)
        mov     esi, edi                    ; copy good value to test
        mov     bx, "90"
        add     eax, edi                    ; check if digit = 9
        jns     FromTop05
@@:
        sub     esi, 100000
        jb      @F
        add     bx, dx                      ; BL + 1, BH - 1
        mov     edi, esi                    ; save good value
        add     eax, 100000
        js      @B

FromTop05:
        mov     bl, bh
        mov     edi, eax
@@:
        mov     [ecx], bl
        add     ecx, 1

Digit04:
        mov     eax, (-90000)
        mov     esi, edi                    ; copy good value to test
        mov     bx, "90"
        add     eax, edi                    ; check if digit = 9
        jns     FromTop04
@@:
        sub     esi, 10000
        jb      @F
        add     bx, dx                      ; BL + 1, BH - 1
        mov     edi, esi                    ; save good value
        add     eax, 10000
        js      @B

FromTop04:
        mov     bl, bh
        mov     edi, eax
@@:
        mov     [ecx], bl
        add     ecx, 1

Digit03:
        mov     eax, (-9000)
        mov     esi, edi                    ; copy good value to test
        mov     bx, "90"
        add     eax, edi                    ; check if digit = 9
        jns     FromTop03
@@:
        sub     esi, 1000
        jb      @F
        add     bx, dx                      ; BL + 1, BH - 1
        mov     edi, esi                    ; save good value
        add     eax, 1000
        js      @B

FromTop03:
        mov     bl, bh
        mov     edi, eax
@@:
        mov     [ecx], bl
        add     ecx, 1

Digit02:
        mov     eax, (-900)
        mov     esi, edi                    ; copy good value to test
        mov     bx, "90"
        add     eax, edi                    ; check if digit = 9
        jns     FromTop02
@@:
        sub     esi, 100
        jb      @F
        add     bx, dx                      ; BL + 1, BH - 1
        mov     edi, esi                    ; save good value
        add     eax, 100
        js      @B

FromTop02:
        mov     bl, bh
        mov     edi, eax
@@:
        mov     [ecx], bl
        add     ecx, 1

Digit01:
        mov     esi, (-90)
        mov     bx, "90"
        add     esi, edi                    ; check if digit = 9
        jns     FromTop01
@@:
        mov     eax, edi                    ; save good value
        sub     edi, 10
        jb      @F
        add     bx, dx                      ; BL + 1, BH - 1
        add     esi, 10
        js      @B

FromTop01:
        mov     bl, bh
        mov     eax, esi                    ; copy good final digit
@@:
        mov     [ecx], bl
        add     ecx, 1

Digit00:
        add     al, "0"
        pop     edx                         ; retrieve original string pointer
        mov     [ecx], ax                   ; add null-termination
        add     ecx, 1                      ; point to null-termination byte
        pop     edi
        mov     eax, ecx
        pop     esi
        sub     eax, edx                    ; get count of chars written
        pop     ebx
        ret

QWtoAconvert endp

; ------------------------------------------------------------------------------

OPTION PROLOGUE:PROLOGUEDEF 
OPTION EPILOGUE:EPILOGUEDEF

